from logging import Logger
from pyspark.sql.session import SparkSession
from datalakebundle.table.create.TableDefinition import TableDefinition


class TablePropertiesSetter:
    def __init__(self, logger: Logger, spark: SparkSession):
        self.__logger = logger
        self.__spark = spark

    def set(self, table_definition: TableDefinition):
        if table_definition.tbl_properties:
            properties: str = ", ".join([f"'{k}'='{v}'" for k, v in table_definition.tbl_properties.items()])

            self.__logger.info(f"Setting TBLPROPERTIES=({properties})")

            self.__spark.sql(f"ALTER TABLE {table_definition.full_table_name} SET TBLPROPERTIES({properties})")



import unittest
from datalakebundle.table.parameters.FieldsResolver import FieldsResolver


class FieldsResolverTest(unittest.TestCase):
    def setUp(self):
        self.__fields_resolver = FieldsResolver()

    def test_basic(self):
        result = self.__fields_resolver.resolve(
            {
                "db_identifier": "mydatabase",
                "table_identifier": "my_table",
            },
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                }
            },
        )

        self.assertEqual(
            {
                "db_identifier": "mydatabase",
                "table_identifier": "my_table",
                "target_path": "/foo/bar/mydatabase/my_table.delta",
            },
            result,
        )

    def test_explicit_overriding_defaults(self):
        result = self.__fields_resolver.resolve(
            {
                "db_identifier": "mydatabase",
                "table_identifier": "my_table",
                "target_path": "/foo/bar/mydatabase/my_table_new2.delta",
            },
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
            },
        )

        self.assertEqual(
            {
                "db_identifier": "mydatabase",
                "table_identifier": "my_table",
                "target_path": "/foo/bar/mydatabase/my_table_new2.delta",
            },
            result,
        )

    def test_infinite_loop_detection(self):
        with self.assertRaises(Exception) as error:
            self.__fields_resolver.resolve(
                {},
                {
                    "target_path": {
                        "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                        "resolver_arguments": ["/foo/bar"],
                    }
                },
            )

        self.assertEqual(
            "Infinite assignment loop detected. Check get_depending_fields() of datalakebundle.test.SimpleTargetPathResolver",
            str(error.exception),
        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable = super-init-not-called
from daipecore.decorator.InputDecorator import InputDecorator
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator


@DecoratedDecorator
class data_frame_saver(InputDecorator):  # pylint: disable = invalid-name
    def __init__(self, *args):
        self._args = args



from logging import Logger
from datalakebundle.delta.DeltaStorage import DeltaStorage
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.write.TablePropertiesSetter import TablePropertiesSetter


class TableRecreator:
    def __init__(
        self,
        logger: Logger,
        delta_storage: DeltaStorage,
        table_properties_setter: TablePropertiesSetter,
    ):
        self.__logger = logger
        self.__delta_storage = delta_storage
        self.__table_properties_setter = table_properties_setter

    def recreate(self, table_definition: TableDefinition):
        self.__logger.info(f"Recreating table {table_definition.full_table_name}")

        self.__delta_storage.recreate_table(table_definition)

        self.__table_properties_setter.set(table_definition)

        self.__logger.info(f"Table {table_definition.full_table_name} successfully recreated")



class IdentifierParser:
    def parse(self, identifier: str) -> dict:
        dot_position = identifier.find(".")

        if dot_position == -1:
            raise Exception("Identifier must meet the following format: {db_identifier}.{table_identifier}")

        return {
            "db_identifier": identifier[:dot_position],
            "table_identifier": identifier[dot_position + 1 :],
            "identifier": identifier,
        }



from typing import Optional
from daipecore.decorator.StringableParameterInterface import StringableParameterInterface


class table_params(StringableParameterInterface):  # pylint: disable = invalid-name
    def __init__(self, identifier: str, param_path_parts: Optional[list] = None):
        self._identifier = identifier
        self._param_path_parts = param_path_parts if param_path_parts else []

    def to_string(self):
        base_path = f'datalakebundle.tables."{self._identifier}".params'

        if self._param_path_parts:
            return "%" + base_path + "." + ".".join(self._param_path_parts) + "%"

        return "%" + base_path + "%"

    def __getattr__(self, item):
        param_path_parts = self._param_path_parts.copy()
        param_path_parts.append(item)

        return table_params(self._identifier, param_path_parts)



import pyspark.sql.types as t


def get_schema():
    return t.StructType(
        [
            t.StructField("name", t.StringType(), True),
            t.StructField("create_time", t.TimestampType(), True),
        ]
    )



from logging import Logger
from pyspark.sql.dataframe import DataFrame
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.schema.SchemaChecker import SchemaChecker
from datalakebundle.table.write.DataWriter import DataWriter
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.create.TableRecreator import TableRecreator


class TableOverwriter:
    def __init__(
        self,
        logger: Logger,
        schema_checker: SchemaChecker,
        table_creator: TableCreator,
        table_recreator: TableRecreator,
        data_writer: DataWriter,
    ):
        self.__logger = logger
        self.__schema_checker = schema_checker
        self.__table_creator = table_creator
        self.__table_recreator = table_recreator
        self.__data_writer = data_writer

    def overwrite(self, result: DataFrame, table_definition: TableDefinition, recreate_table: bool, options: dict):
        output_table_name = table_definition.full_table_name

        self.__schema_checker.check(result, table_definition)

        self.__logger.info(f"Data to be written into table: {output_table_name}")

        if recreate_table:
            self.__table_recreator.recreate(table_definition)
        else:
            self.__table_creator.create_if_not_exists(table_definition)
            self.__schema_checker.check_metadata(table_definition)

        self.__logger.info(f"Writing data to table: {output_table_name}")

        self.__data_writer.overwrite(result, output_table_name, table_definition.partition_by, options)

        self.__logger.info(f"Data successfully written to: {output_table_name}")



from daipecore.lineage.OutputDecoratorInterface import OutputDecoratorInterface


class TableWriter(OutputDecoratorInterface):
    def __init__(self, full_table_name: str, mode: str):
        self.__full_table_name = full_table_name
        self.__mode = mode

    @property
    def full_table_name(self):
        return self.__full_table_name

    @property
    def mode(self):
        return self.__mode

    @property
    def identifier(self):
        return self.__full_table_name



import io
import os
from contextlib import redirect_stdout
import pandas as pd
from datalakebundle.notebook.decorator.data_frame_loader import data_frame_loader
from datalakebundle.notebook.decorator.transformation import transformation
from datalakebundle.notebook.decorator.data_frame_saver import data_frame_saver

os.environ["APP_ENV"] = "test"

# pylint: disable = invalid-name
printed = """   a  b
0  0  2
1  1  3
"""


@data_frame_loader()
def load_data():
    return pd.DataFrame({"a": [0, 1]})


@transformation()
def load_data2():
    return pd.DataFrame({"b": [2, 3]})


@transformation(load_data, load_data2, check_duplicate_columns=False)
def concat(df1: pd.DataFrame, df2: pd.DataFrame):
    assert df1["a"].values.tolist() == [0, 1]
    assert df2["b"].values.tolist() == [2, 3]

    return pd.concat([df1, df2], axis=1)


@data_frame_saver(concat)
def save(result: pd.DataFrame):
    assert result.shape == (2, 2)


def get_list(df: pd.DataFrame, column: str):
    return df[column].values.tolist()


if __name__ == "__main__":
    assert isinstance(load_data, data_frame_loader)
    assert get_list(load_data.result, "a") == [0, 1]
    assert get_list(globals()["load_data_df"], "a") == [0, 1]

    assert isinstance(load_data2, transformation)
    assert get_list(load_data2.result, "b") == [2, 3]
    assert get_list(globals()["load_data2_df"], "b") == [2, 3]

    assert isinstance(concat, transformation)
    assert get_list(concat.result, "a") == [0, 1]
    assert get_list(concat.result, "b") == [2, 3]
    assert get_list(globals()["concat_df"], "a") == [0, 1]
    assert get_list(globals()["concat_df"], "b") == [2, 3]

    assert isinstance(save, data_frame_saver)
    assert save.result is None
    assert "save_df" not in globals()

    with redirect_stdout(io.StringIO()) as f:

        @transformation(load_data, load_data2, check_duplicate_columns=False, display=True)
        def concat2(df1: pd.DataFrame, df2: pd.DataFrame):
            assert df1["a"].values.tolist() == [0, 1]
            assert df2["b"].values.tolist() == [2, 3]

            return pd.concat([df1, df2], axis=1)

    assert printed == f.getvalue()



import _ast
from daipecore.lineage.DecoratorParserInterface import DecoratorParserInterface
from datalakebundle.notebook.lineage.TableWriter import TableWriter


class TableWriterParser(DecoratorParserInterface):
    def __init__(self, name: str, mode: str):
        self.__name = name
        self.__mode = mode

    def parse(self, decorator: _ast.Call):
        if hasattr(_ast, "Str"):
            arg: _ast.Str = decorator.args[0]  # pyre-ignore[11]
            return TableWriter(arg.s, self.__mode)  # pyre-ignore[16]

        arg: _ast.Constant = decorator.args[0]  # pyre-ignore[9]
        return TableWriter(arg.value, self.__mode)

    def get_name(self) -> str:
        return self.__name



import unittest
from injecta.testing.services_tester import test_services
from pyfonycore.bootstrap import bootstrapped_container
from datalakebundle.table.parameters.TableParameters import TableParameters
from datalakebundle.table.parameters.TableParametersManager import TableParametersManager


class DataLakeBundleTest(unittest.TestCase):
    def setUp(self):
        self._container = bootstrapped_container.init("test")
        self.max_diff = 8000 * 8

    def test_init(self):
        test_services(self._container)

    def test_table_parameters(self):
        table_parameters_manager: TableParametersManager = self._container.get(TableParametersManager)

        my_table_parameters = table_parameters_manager.get_or_parse("test_{identifier}", "mydatabase_e.my_table")
        another_table_parameters = table_parameters_manager.get_or_parse("test_{identifier}", "mydatabase_p.another_table")

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase_e",
                    "table_identifier": "my_table",
                    "identifier": "mydatabase_e.my_table",
                    "db_name": "test_mydatabase_e",
                    "table_name": "my_table",
                    "encrypted": True,
                    "db_identifier_base": "mydatabase",
                    "target_path": "/foo/bar/mydatabase/encrypted/my_table.delta",
                }
            ),
            my_table_parameters,
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase_p",
                    "table_identifier": "another_table",
                    "identifier": "mydatabase_p.another_table",
                    "db_name": "test_mydatabase_p",
                    "table_name": "another_table",
                    "encrypted": False,
                    "db_identifier_base": "mydatabase",
                    "target_path": "/foo/bar/mydatabase/plain/another_table.delta",
                }
            ),
            another_table_parameters,
        )


if __name__ == "__main__":
    unittest.main()



import pyspark.sql.types as t

df_schema_array2 = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(
                t.ArrayType(
                    t.StructType(
                        [
                            t.StructField("FIELD4", t.IntegerType()),
                            t.StructField("FIELD5", t.IntegerType()),
                        ]
                    )
                )
            ),
        ),
    ],
)



import pyspark.sql.types as t

expected_schema_nullable3 = t.StructType(
    [
        t.StructField("FIELD1_nullable", t.IntegerType(), nullable=True),
        t.StructField("nullable", t.StringType(), nullable=False),
        t.StructField("not_nullable", t.StringType()),
        t.StructField(
            "STRUCT1",
            t.StructType(
                [
                    t.StructField("NESTED_FIELD1", t.StringType(), nullable=True),
                    t.StructField(
                        "['nullable']",
                        t.StructType(
                            [
                                t.StructField("NESTED_FIELD2", t.StringType(), nullable=False),
                            ],
                        ),
                    ),
                ],
            ),
        ),
    ],
)



import pprint
import re
from deepdiff import DeepDiff
from pyspark.sql.types import StructType


class DiffGenerator:
    def generate(self, df_schema: StructType, schema: StructType):
        def remove_metadata(json_schema):
            for field in json_schema["fields"]:
                field["metadata"] = {}

            return json_schema

        expected_schema = remove_metadata(schema.jsonValue())
        df_schema = remove_metadata(df_schema.jsonValue())

        exclude_nullable = re.compile(r"\['nullable'\]")
        ddiff = DeepDiff(expected_schema, df_schema, ignore_string_case=True, ignore_order=True, exclude_regex_paths=[exclude_nullable])

        result = []
        if ddiff:

            if "values_changed" in ddiff:
                result.extend(self.__get_changed(expected_schema, ddiff["values_changed"]))

            if "type_changes" in ddiff:
                result.extend(self.__get_changed(expected_schema, ddiff["type_changes"], is_values=False))

            if "iterable_item_added" in ddiff:
                result.extend(self.__get_iterable_item(expected_schema, ddiff["iterable_item_added"], "unexpected field"))

            if "iterable_item_removed" in ddiff:
                result.extend(self.__get_iterable_item(expected_schema, ddiff["iterable_item_removed"], "missing field"))

            if "dictionary_item_added" in ddiff:
                result.extend(self.__get_dictionary_item(expected_schema, ddiff["dictionary_item_added"], "Unexpected field"))

            if "dictionary_item_removed" in ddiff:
                result.extend(self.__get_dictionary_item(expected_schema, ddiff["dictionary_item_removed"], "Missing field"))

        return result

    def __get_field_names(self, current: dict, chnks: list):
        if chnks:
            chunk = chnks[0]
            if len(chunk) == 1 and len(chnks) == 1:
                return
            if chunk[0] == "elementType":
                yield "array"
                yield from self.__get_field_names(current["elementType"], chnks[1:])
            else:
                new = current[chunk[0]][int(chunk[1])]
                yield new["name"]
                yield from self.__get_field_names(new["type"], chnks[1:])

    def __get_field_path(self, expected_schema: dict, key: str, cut_last_chunk: bool = False):
        path = key[4:].replace("'", "").replace("[", "").split("]")[:-1]

        def chunks(lst: list, n: int):
            if not lst:
                return
            if lst[0] in ["containsNull", "elementType"]:
                yield [lst[0]]
                yield from chunks(lst[1:], n)
            else:
                yield lst[0:n]
                yield from chunks(lst[n:], n)

        chnks = list(chunks(path, 3))

        if cut_last_chunk:
            chnks.pop()

        field_names = self.__get_field_names(expected_schema, chnks)
        return ".".join(field_names), path[-1]

    def __get_changed(self, expected_schema: dict, values_changed: dict, is_values: bool = True):
        result = []

        for key, val in values_changed.items():
            field_path, last_field = self.__get_field_path(expected_schema, key)

            old_value = val["old_value"]
            new_value = val["new_value"]

            if is_values:
                old_value, new_value, last_field = self.__values_changed(old_value, new_value, last_field)
            else:
                old_value, new_value, last_field = self.__types_changed(old_value, new_value, last_field)

            result.append(f"{field_path}{last_field} changed from {old_value} to {new_value}")

        return result

    def __values_changed(self, old_value, new_value, last_field):
        if isinstance(old_value, str) and old_value not in ["struct", "array"]:
            old_value = old_value.upper()
        if isinstance(new_value, str) and new_value not in ["struct", "array"]:
            new_value = new_value.upper()
        if last_field.isnumeric():
            last_field = ""
        else:
            last_field = f"['{last_field}']"

        if isinstance(old_value, dict):
            old_value = pprint.pformat(old_value, indent=4)
        if isinstance(new_value, dict):
            new_value = pprint.pformat(new_value, indent=4)

        return old_value, new_value, last_field

    def __types_changed(self, old_value, new_value, last_field):
        if isinstance(old_value, dict):
            if "name" in old_value:
                old_value = f"struct ({old_value['name']})"
            else:
                old_value = "array"
        if isinstance(new_value, dict):
            if "name" in new_value:
                new_value = f"struct ({new_value['name']})"
            else:
                new_value = "array"

        return old_value, new_value, f"['{last_field}']"

    def __get_iterable_item(self, expected_schema: dict, iterable_items: dict, message: str):
        result = []

        for key, val in iterable_items.items():
            field_path, _ = self.__get_field_path(expected_schema, key, True)
            if not field_path:
                field_path = "root"
            result.append(f'{field_path} {message}: {val["name"].upper()}')
        return result

    def __get_dictionary_item(self, expected_schema: dict, dictionary_items: list, message: str):
        result = []

        for item in dictionary_items:
            field_path, last_field = self.__get_field_path(expected_schema, item)

            if not field_path:
                field_path = "root"

            if last_field == "fields":
                last_field = ".struct"
            elif last_field in ["elementType", "containsNull"]:
                last_field = ".array"
            else:
                last_field = f"[{last_field}]"

            result.append(f"{message} {field_path}{last_field}")
        return list(dict.fromkeys(result))



class TableNames:
    def __init__(self, db_identifier: str, db_name: str, table_identifier: str, table_name: str):
        self.__db_identifier = db_identifier
        self.__db_name = db_name
        self.__table_identifier = table_identifier
        self.__table_name = table_name

    @property
    def db_identifier(self):
        return self.__db_identifier

    @property
    def db_name(self):
        return self.__db_name

    @property
    def table_identifier(self):
        return self.__table_identifier

    @property
    def table_name(self):
        return self.__table_name

    @property
    def identifier(self):
        return self.__db_identifier + "." + self.__table_identifier

    @property
    def full_table_name(self):
        return self.__db_name + "." + self.__table_name

    def to_dict(self):
        return dict(
            identifier=self.identifier,
            db_identifier=self.__db_identifier,
            db_name=self.__db_name,
            table_identifier=self.__table_identifier,
            table_name=self.__table_name,
        )



from datalakebundle.table.identifier.IdentifierParser import IdentifierParser
from datalakebundle.table.name.TableNamesParser import TableNamesParser


class TableNamesPreparer:
    def __init__(self, identifier_parser: IdentifierParser, table_names_parser: TableNamesParser):
        self.__identifier_parser = identifier_parser
        self.__table_names_parser = table_names_parser

    def prepare(self, table_name_template: str, identifier: str):
        identifiers = self.__identifier_parser.parse(identifier)

        return self.__table_names_parser.parse(table_name_template, identifiers)



from argparse import ArgumentParser, Namespace
from logging import Logger
from time import sleep
from consolebundle.ConsoleCommand import ConsoleCommand
from pysparkbundle.filesystem.FilesystemInterface import FilesystemInterface
from datalakebundle.table.create.TableDefinitionFactory import TableDefinitionFactory
from datalakebundle.table.create.TableRecreator import TableRecreator
from datalakebundle.table.class_ import table_schema_loader


class TableRecreatorCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        filesystem: FilesystemInterface,
        table_definition_factory: TableDefinitionFactory,
        table_recreator: TableRecreator,
    ):
        self.__logger = logger
        self.__filesystem = filesystem
        self.__table_definition_factory = table_definition_factory
        self.__table_recreator = table_recreator

    def get_command(self) -> str:
        return "datalake:table:recreate"

    def get_description(self):
        return "Recreates a table based on it's identifier and table class"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")
        argument_parser.add_argument(dest="table_schema_path", help="Table class path [module_path].[class_name]")
        argument_parser.add_argument(
            "-s",
            "--skip-countdown",
            action="store_true",
            help="Skip data deletion countdown",
        )

    def run(self, input_args: Namespace):
        table_definition = self.__table_definition_factory.create_from_table_schema(
            input_args.identifier, table_schema_loader.load(input_args.table_schema_path)
        )

        if input_args.skip_countdown is False and self.__filesystem.exists(table_definition.target_path):
            self.__logger.info("Use the --skip-countdown switch to delete existing data immediately")

            countdown = 10

            for i in range(countdown):
                self.__logger.warning(f"Table data will be deleted in {countdown - i}s. Use CTRL+C to cancel.")
                sleep(1)

        self.__table_recreator.recreate(table_definition)



from pyspark.sql.types import StructType


class UpsertQueryCreator:
    def create(self, full_table_name: str, schema: StructType, primary_key: list, temp_source_table: str) -> str:
        conditions = []
        updates = []
        columns_to_update = set(schema.fieldNames()) - set(primary_key)

        for primary_key_column in primary_key:
            conditions.append(f"source.`{primary_key_column}` = target.`{primary_key_column}`")

        for col in columns_to_update:
            updates.append(f"target.`{col}` = source.`{col}`")

        query = (
            f"MERGE INTO {full_table_name} AS target\n"
            f"USING {temp_source_table} AS source\n"
            f"ON {' AND '.join(conditions)}\n"
            f"{{matched_clause}}"
            f"WHEN NOT MATCHED THEN INSERT *\n"
        )

        if len(updates) > 0:
            query = query.format(matched_clause=f"WHEN MATCHED THEN UPDATE SET {', '.join(updates)}\n")
        else:
            query = query.format(matched_clause="")

        return query



from pyfonybundles.Bundle import Bundle
from datalakebundle.table.identifier.Expression import Expression

Expression.yaml_loader.add_constructor(Expression.yaml_tag, Expression.from_yaml)


class DataLakeBundle(Bundle):
    def modify_raw_config(self, raw_config: dict) -> dict:
        table_defaults = raw_config["parameters"]["datalakebundle"]["table"]["defaults"]
        if table_defaults:
            self.__check_defaults(table_defaults)

        return raw_config

    def __check_defaults(self, table_defaults: dict):
        for field in ["identifier", "db_identifier", "db_name", "table_identifier", "table_name"]:
            if field in table_defaults:
                raise Exception(f"datalakebundle.table.defaults.{field} parameter must not be explicitly defined")



# pylint: disable = super-init-not-called
from pyspark.sql import SparkSession


class DummySparkFactory:
    def create(self):
        class DummySparkSession(SparkSession):
            def __init__(self, *args):
                pass

        return DummySparkSession()



from datalakebundle.table.identifier.Expression import Expression
from datalakebundle.table.identifier.fill_template import fill_template


def evaluate(val, raw_table_parameters: dict):
    if isinstance(val, Expression):
        return val.evaluate(raw_table_parameters)

    if isinstance(val, str):
        return fill_template(val, raw_table_parameters)

    return val



import sys
from argparse import Namespace, ArgumentParser
from logging import Logger
from pyspark.sql import SparkSession
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.parameters.TableParametersManager import TableParametersManager
from datalakebundle.table.name.TableNameTemplateGetter import TableNameTemplateGetter


class TableOptimizerCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
        table_name_template_getter: TableNameTemplateGetter,
        table_parameters_manager: TableParametersManager,
        table_existence_checker: TableExistenceChecker,
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_name_template_getter = table_name_template_getter
        self.__table_parameters_manager = table_parameters_manager
        self.__table_existence_checker = table_existence_checker

    def get_command(self) -> str:
        return "datalake:table:optimize"

    def get_description(self):
        return "OPTIMIZEs given table"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")

    def run(self, input_args: Namespace):
        table_name_template = self.__table_name_template_getter.get_template_for_write()
        table_parameters = self.__table_parameters_manager.get_or_parse(table_name_template, input_args.identifier)

        if not self.__table_existence_checker.table_exists(table_parameters.db_name, table_parameters.table_name):
            self.__logger.error(f"Hive table {table_parameters.full_table_name} does NOT exists")
            sys.exit(1)

        self.__logger.info(f"Running OPTIMIZE {table_parameters.full_table_name}")

        self.__spark.sql(f"OPTIMIZE {table_parameters.full_table_name}")

        self.__logger.info(f"OPTIMIZE for {table_parameters.full_table_name} completed")



from daipecore.lineage.argument.DecoratorInputFunctionInterface import DecoratorInputFunctionInterface


class ReadTable(DecoratorInputFunctionInterface):
    def __init__(self, full_table_name: str):
        self.__full_table_name = full_table_name

    @property
    def full_table_name(self):
        return self.__full_table_name

    @property
    def identifier(self):
        return self.__full_table_name



from logging import Logger
from pyspark.sql.session import SparkSession
from pysparkbundle.filesystem.FilesystemInterface import FilesystemInterface
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.UnknownTableException import UnknownTableException
from datalakebundle.table.parameters.TableParameters import TableParameters


class TableDeleter:
    def __init__(
        self,
        logger: Logger,
        filesystem: FilesystemInterface,
        table_existence_checker: TableExistenceChecker,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__filesystem = filesystem
        self.__table_existence_checker = table_existence_checker
        self.__spark = spark

    def delete_including_data(self, table_parameters: TableParameters):
        self.__logger.info(f"Hive table to be deleted: {table_parameters.full_table_name}")
        self.__logger.warning(f"Data lake files to be deleted: {table_parameters.target_path}")

        if self.__table_existence_checker.table_exists(table_parameters.db_name, table_parameters.table_name) is False:
            raise UnknownTableException(f"Table {table_parameters.full_table_name} does not exist in Hive")

        self.__spark.sql(f"DROP TABLE {table_parameters.full_table_name}")
        self.__filesystem.delete(table_parameters.target_path, True)

        self.__logger.info(f"Hive table {table_parameters.full_table_name} and data deleted")



# pylint: disable = unused-import

from daipecore.decorator.notebook_function import notebook_function
from datalakebundle.notebook.decorator.transformation import transformation
from datalakebundle.notebook.decorator.data_frame_loader import data_frame_loader
from datalakebundle.notebook.decorator.data_frame_saver import data_frame_saver
from pysparkbundle.csv.csv_reader import read_csv
from pysparkbundle.csv.csv_append import csv_append
from pysparkbundle.csv.csv_overwrite import csv_overwrite
from pysparkbundle.csv.csv_write_ignore import csv_write_ignore
from pysparkbundle.csv.csv_write_errorifexists import csv_write_errorifexists
from pysparkbundle.delta.delta_reader import read_delta
from pysparkbundle.delta.delta_append import delta_append
from pysparkbundle.delta.delta_overwrite import delta_overwrite
from pysparkbundle.delta.delta_write_ignore import delta_write_ignore
from pysparkbundle.delta.delta_write_errorifexists import delta_write_errorifexists
from pysparkbundle.json.json_reader import read_json
from pysparkbundle.json.json_append import json_append
from pysparkbundle.json.json_overwrite import json_overwrite
from pysparkbundle.json.json_write_ignore import json_write_ignore
from pysparkbundle.json.json_write_errorifexists import json_write_errorifexists
from pysparkbundle.parquet.parquet_reader import read_parquet
from pysparkbundle.parquet.parquet_append import parquet_append
from pysparkbundle.parquet.parquet_overwrite import parquet_overwrite
from pysparkbundle.parquet.parquet_write_ignore import parquet_write_ignore
from pysparkbundle.parquet.parquet_write_errorifexists import parquet_write_errorifexists
from datalakebundle.table.parameters.table_params import table_params
from datalakebundle.table.write.table_append import table_append
from datalakebundle.table.write.table_overwrite import table_overwrite
from datalakebundle.table.write.table_upsert import table_upsert
from datalakebundle.table.read.table_reader import read_table
from datalakebundle.table.schema.TableSchema import TableSchema



from logging import Logger
from datalakebundle.delta.DeltaStorage import DeltaStorage
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.write.TablePropertiesSetter import TablePropertiesSetter


class TableCreator:
    def __init__(
        self,
        logger: Logger,
        delta_storage: DeltaStorage,
        table_existence_checker: TableExistenceChecker,
        table_properties_setter: TablePropertiesSetter,
    ):
        self.__logger = logger
        self.__delta_storage = delta_storage
        self.__table_existence_checker = table_existence_checker
        self.__table_properties_setter = table_properties_setter

    def create(self, table_definition: TableDefinition):
        self.__logger.info(f"Creating new table {table_definition.full_table_name} for {table_definition.target_path}")

        self.__delta_storage.create_table(table_definition)

        self.__table_properties_setter.set(table_definition)

        self.__logger.info(f"Table {table_definition.full_table_name} successfully created")

    def create_if_not_exists(self, table_definition: TableDefinition):
        if self.__table_existence_checker.table_exists(table_definition.db_name, table_definition.table_name):
            self.__logger.info(f"Table {table_definition.full_table_name} already exists, creation skipped")
            return

        self.create(table_definition)



from box import Box
from datalakebundle.table.identifier.ValueResolverInterface import ValueResolverInterface


class TargetPathResolver(ValueResolverInterface):
    def __init__(self, base_path: str):
        self.__base_path = base_path

    def resolve(self, raw_table_parameters: Box):
        encrypted_string = "encrypted" if raw_table_parameters.encrypted is True else "plain"

        return (
            self.__base_path
            + "/"
            + raw_table_parameters.db_identifier_base
            + "/"
            + encrypted_string
            + "/"
            + raw_table_parameters.table_identifier
            + ".delta"
        )



from logging import Logger
from pyspark.sql import DataFrame
from pyspark.sql.types import StructType
from datalakebundle.table.parameters.TableParameters import TableParameters
from datalakebundle.table.parameters.TableParametersManager import TableParametersManager
from datalakebundle.table.name.TableNameTemplateGetter import TableNameTemplateGetter
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.schema.TableSchemaGenerator import TableSchemaGenerator
from datalakebundle.table.schema.TableSchema import TableSchema


class TableDefinitionFactory:
    def __init__(
        self,
        logger: Logger,
        table_name_template_getter: TableNameTemplateGetter,
        table_parameters_manager: TableParametersManager,
        table_schema_generator: TableSchemaGenerator,
    ):
        self.__logger = logger
        self.__table_name_template_getter = table_name_template_getter
        self.__table_parameters_manager = table_parameters_manager
        self.__table_schema_generator = table_schema_generator

    def create_from_table_schema(self, identifier: str, table_schema: TableSchema):
        table_name_template = self.__table_name_template_getter.get_template_for_write()
        table_parameters = self.__table_parameters_manager.get_or_parse(table_name_template, identifier)

        schema = StructType(table_schema.fields)
        primary_key = [table_schema.primary_key] if isinstance(table_schema.primary_key, str) else table_schema.primary_key

        if hasattr(table_schema, "partition_by"):
            partition_by = [table_schema.partition_by] if isinstance(table_schema.partition_by, str) else table_schema.partition_by
        else:
            partition_by = []

        tbl_properties = table_schema.tbl_properties

        return self.__create(table_parameters, schema, primary_key, partition_by, tbl_properties)

    def create_from_dataframe(self, identifier: str, df: DataFrame, decorator_name: str):
        self.__logger.warning(
            f"No explicit schema provided, using dataframe schema instead. "
            f"You can define schema as:\n\n{self.__table_schema_generator.generate(df.schema)}\n"
            f'Usage: @dp.{decorator_name}("{identifier}", get_schema())'
        )
        table_name_template = self.__table_name_template_getter.get_template_for_write()
        table_parameters = self.__table_parameters_manager.get_or_parse(table_name_template, identifier)

        return self.__create(table_parameters, df.schema, [], [], {})

    def __create(self, table_parameters: TableParameters, schema: StructType, primary_key: list, partition_by: list, tbl_properties: dict):
        return TableDefinition(
            table_parameters.db_name,
            table_parameters.table_name,
            schema,
            primary_key,
            partition_by,
            table_parameters.target_path,
            tbl_properties,
        )



from injecta.module import attribute_loader
from injecta.service.parser.DTypeResolver import DTypeResolver
from datalakebundle.table.identifier.ValueResolverInterface import ValueResolverInterface


class ValueResolverFactory:
    def __init__(self):
        self.__d_type_resolver = DTypeResolver()

    def create(self, val) -> ValueResolverInterface:
        resolver_type = self.__d_type_resolver.resolve(val["resolver_class"])
        resolver_class = attribute_loader.load(resolver_type.module_name, resolver_type.class_name)
        resolver_arguments = val["resolver_arguments"] if "resolver_arguments" in val else []

        return resolver_class(*resolver_arguments)



import pyspark.sql.types as t

df_schema2 = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.StringType()),
        t.StructField("FIELD3", t.StringType()),
        t.StructField(
            "STRUCT1",
            t.StructType(
                [
                    t.StructField("NESTED_FIELD1", t.StringType()),
                    t.StructField(
                        "STRUCT2",
                        t.StructType(
                            [
                                t.StructField("NESTED_FIELD2", t.StringType()),
                            ],
                        ),
                    ),
                ],
            ),
        ),
    ],
)



import pyspark.sql.types as t

expected_schema6 = t.StructType(
    [
        t.StructField(
            "FIELD6",
            t.StructField(
                "STRUCT1",
                t.StructType(
                    [
                        t.StructField("NESTED_FIELD0", t.StringType()),
                    ],
                ),
            ),
        ),
    ]
)



import unittest
from datalakebundle.table.schema.TableSchema import TableSchema
from pyspark.sql import types as t


class TableSchemaTest(unittest.TestCase):
    def test_to_json(self):
        schema = self.__get_basic_schema()
        schema_json = schema.jsonValue()

        self.assertEqual({"name": "name", "type": "string", "nullable": False, "metadata": {}}, schema_json["fields"][0])
        self.assertEqual({"name": "date", "type": "date", "nullable": False, "metadata": {}}, schema_json["fields"][1])
        self.assertEqual({"name": "visits", "type": "integer", "nullable": True, "metadata": {}}, schema_json["fields"][2])

        # "struct" type is important for spark data loaders to work properly
        self.assertEqual("struct", schema_json["type"])

    def __get_basic_schema(self):
        return TableSchema(
            [
                t.StructField("name", t.StringType(), False),
                t.StructField("date", t.DateType(), False),
                t.StructField("visits", t.IntegerType(), True),
            ],
            primary_key=["name", "date"],
            partition_by=["name"],
        )


if __name__ == "__main__":
    unittest.main()



import sys
from argparse import ArgumentParser, Namespace
from logging import Logger
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.class_ import table_schema_loader
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.create.TableDefinitionFactory import TableDefinitionFactory


class TableCreatorCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_definition_factory: TableDefinitionFactory,
        table_existence_checker: TableExistenceChecker,
        table_creator: TableCreator,
    ):
        self.__logger = logger
        self.__table_definition_factory = table_definition_factory
        self.__table_existence_checker = table_existence_checker
        self.__table_creator = table_creator

    def get_command(self) -> str:
        return "datalake:table:create"

    def get_description(self):
        return "Creates a metastore table based on it's identifier and table class"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")
        argument_parser.add_argument(dest="table_schema_path", help="Table class path [module_path].[class_name]")

    def run(self, input_args: Namespace):
        table_definition = self.__table_definition_factory.create_from_table_schema(
            input_args.identifier, table_schema_loader.load(input_args.table_schema_path)
        )

        if self.__table_existence_checker.table_exists(table_definition.db_name, table_definition.table_name):
            self.__logger.error(f"Hive table {table_definition.full_table_name} already exists")
            sys.exit(1)

        self.__table_creator.create(table_definition)



from daipecore.lineage.argument.ArgumentMappingInterface import ArgumentMappingInterface
from datalakebundle.notebook.lineage.argument.ReadTable import ReadTable
from datalakebundle.notebook.lineage.argument.TableParams import TableParams


class ArgumentMapping(ArgumentMappingInterface):
    def get_mapping(self):
        return {
            "table_params": TableParams,
            "read_table": ReadTable,
        }



import unittest
from types import FunctionType


class DecoratorsTest(unittest.TestCase):
    def test_basic(self):
        from datalakebundle.notebook.decorator.tests.decorators_test import (  # pylint: disable = import-outside-toplevel
            load_data2,
            get_list,
        )

        result = load_data2()

        self.assertIsInstance(load_data2, FunctionType)
        self.assertEqual([2, 3], get_list(result, "b"))

    def test_error(self):
        with self.assertRaises(Exception) as error:
            from datalakebundle.notebook.decorator.tests.decorators_fixture import (  # pylint: disable = import-outside-toplevel, unused-import
                load_data3,
            )

        self.assertEqual("Use @data_frame_loader() instead of @data_frame_loader please", str(error.exception))


if __name__ == "__main__":
    unittest.main()



import unittest
from datalakebundle.table.identifier.fill_template import fill_template


class fill_template_test(unittest.TestCase):  # pylint: disable = invalid-name
    def test_basic(self):
        values = {"name": "Jiri", "age": 34}

        result = fill_template("Hello {name}, I am {age} years old", values)

        self.assertEqual("Hello Jiri, I am 34 years old", result)

    def test_missing_value(self):
        with self.assertRaises(Exception) as error:
            values = {"name": "Jiri"}

            fill_template("Hello {name}, I am {age} years old", values)

        self.assertEqual("Value for placeholder {age} not defined", str(error.exception))


if __name__ == "__main__":
    unittest.main()



from datalakebundle.table.identifier.fill_template import fill_template
from datalakebundle.table.name.TableNames import TableNames


class TableNamesParser:
    def parse(self, table_name_template: str, identifiers: dict):
        full_table_name = fill_template(table_name_template, identifiers)
        dot_position = full_table_name.find(".")

        if dot_position == -1:
            raise Exception("Table name must meet the following format: {db_name}.{table_name}")

        return TableNames(
            identifiers["db_identifier"],
            full_table_name[:dot_position],
            identifiers["table_identifier"],
            full_table_name[dot_position + 1 :],
        )



# pylint: disable = super-init-not-called
from typing import Optional
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from daipecore.decorator.OutputDecorator import OutputDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import DataFrame
from datalakebundle.table.create.TableDefinitionFactory import TableDefinitionFactory
from datalakebundle.table.schema.TableSchema import TableSchema
from datalakebundle.table.write.TableOverwriter import TableOverwriter


@DecoratedDecorator
class table_overwrite(OutputDecorator):  # pylint: disable = invalid-name
    def __init__(
        self, identifier: str, table_schema: Optional[TableSchema] = None, recreate_table: bool = False, options: Optional[dict] = None
    ):
        self.__identifier = identifier
        self.__table_schema = table_schema
        self.__recreate_table = recreate_table
        self.__options = options or {}

    def process_result(self, result: DataFrame, container: ContainerInterface):
        table_definition_factory: TableDefinitionFactory = container.get(TableDefinitionFactory)
        table_overwriter: TableOverwriter = container.get(TableOverwriter)

        if self.__table_schema:
            table_definition = table_definition_factory.create_from_table_schema(self.__identifier, self.__table_schema)
        else:
            table_definition = table_definition_factory.create_from_dataframe(self.__identifier, result, self.__class__.__name__)

        table_overwriter.overwrite(result, table_definition, self.__recreate_table, self.__options)



from typing import Callable, List
from box import Box
from datalakebundle.table.parameters.TableParameters import TableParameters
from datalakebundle.table.parameters.TableParametersParser import TableParametersParser


class TableParametersManager:

    __table_parameters: List[TableParameters] = []

    def __init__(
        self,
        raw_table_parameters: Box,
        table_defaults: Box,
        table_parameters_parser: TableParametersParser,
    ):
        self.__table_defaults = table_defaults.to_dict() if table_defaults else {}
        self.__table_parameters_parser = table_parameters_parser

        raw_table_parameters = raw_table_parameters or Box({})

        self.__table_parameters = [
            table_parameters_parser.parse(identifier, self.__table_defaults, explicit_parameters or {})
            for identifier, explicit_parameters in raw_table_parameters.to_dict().items()
        ]

    def exists(self, identifier: str) -> bool:
        for table_parameters in self.__table_parameters:
            if table_parameters.identifier == identifier:
                return True

        return False

    def get_or_parse(self, table_name_template: str, identifier: str) -> TableParameters:
        for table_parameters in self.__table_parameters:
            if table_parameters.identifier == identifier:
                return table_parameters

        # no explicitly defined table parameters found, parse the basic parameters from identifier only
        return self.__table_parameters_parser.parse(table_name_template, identifier, self.__table_defaults)

    def get_all(self):
        return self.__table_parameters

    def get_by_filter(self, filter_function: Callable):
        return list(filter(filter_function, self.__table_parameters))

    def get_with_attribute(self, attr_name: str):
        return self.get_by_filter(lambda table_parameters: hasattr(table_parameters, attr_name))



import os
from datalakebundle.notebook.decorator.data_frame_loader import data_frame_loader
from datalakebundle.notebook.decorator.transformation import transformation
from datalakebundle.notebook.decorator.data_frame_saver import data_frame_saver

os.environ["APP_ENV"] = "test"


@data_frame_loader()
def load_data():
    return 155


@data_frame_loader()
def load_data2():
    return 145


@transformation(load_data, load_data2, check_duplicate_columns=False)
def sumup(police_number: int, something: int):
    assert police_number == 155
    assert something == 145

    return 155 + 145


@data_frame_loader
def load_data3():
    return 111


@data_frame_saver(sumup)
def save(result: int):
    assert result == 300



import types
from injecta.module import attribute_loader


def load(table_schema_path: str):
    last_dot_pos = table_schema_path.rfind(".")

    if not last_dot_pos:
        raise Exception(f"Invalid class path: {table_schema_path}")

    module_name = table_schema_path[:last_dot_pos]
    attribute_name = table_schema_path[last_dot_pos + 1 :]

    schema = attribute_loader.load(module_name, attribute_name)

    if isinstance(schema, types.FunctionType):
        return schema()

    return schema



# pylint: disable = super-init-not-called
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from injecta.container.ContainerInterface import ContainerInterface
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator
from pysparkbundle.dataframe.DataFrameShowMethodInterface import DataFrameShowMethodInterface


@DecoratedDecorator
class data_frame_loader(DataFrameReturningDecorator):  # pylint: disable = invalid-name
    def __init__(self, *args, display=False):
        self._args = args
        self._display = display

    def after_execution(self, container: ContainerInterface):
        self._set_global_dataframe()

        if self._display and container.get_parameters().datalakebundle.notebook.display.enabled is True:
            data_frame_show_method: DataFrameShowMethodInterface = container.get("pysparkbundle.dataframe.show_method")
            data_frame_show_method.show(self._result)



import unittest
from pyfonycore.bootstrap import bootstrapped_container
from datalakebundle.table.parameters.TableParameters import TableParameters
from datalakebundle.table.parameters.TableParametersParser import TableParametersParser


class TableParametersParserTest(unittest.TestCase):
    __table_parameters_parser: TableParametersParser

    def setUp(self):
        container = bootstrapped_container.init("test")

        self.__table_parameters_parser = container.get(TableParametersParser)

    def test_basic(self):
        result = self.__table_parameters_parser.parse(
            "test_{identifier}",
            "mydatabase.my_table",
            {},
            {
                "target_path": "/foo/bar/mydatabase/my_table.delta",
            },
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                }
            ),
            result,
        )

    def test_defaults_only(self):
        result = self.__table_parameters_parser.parse(
            "test_{identifier}",
            "mydatabase.my_table",
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
            },
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                }
            ),
            result,
        )

    def test_explicit_overriding_defaults(self):
        result = self.__table_parameters_parser.parse(
            "test_{identifier}",
            "mydatabase.my_table",
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
            },
            {},
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                }
            ),
            result,
        )

    def test_explicit_overriding_db_name(self):
        result = self.__table_parameters_parser.parse(
            "test_{identifier}",
            "mydatabase.my_table",
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
            },
            {
                "db_name": "prod_mydatabase",
            },
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "db_name": "prod_mydatabase",
                    "table_name": "my_table",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                }
            ),
            result,
        )

    def test_explicit_overriding_table_name(self):
        result = self.__table_parameters_parser.parse(
            "test_{identifier}",
            "mydatabase.my_table",
            {
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
            },
            {
                "table_name": "prod_my_table",
            },
        )

        self.assertEqual(
            TableParameters(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "prod_my_table",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                }
            ),
            result,
        )


if __name__ == "__main__":
    unittest.main()



from logging import Logger
from daipecore.function.input_decorator_function import input_decorator_function
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import SparkSession
from datalakebundle.table.parameters.TableParametersManager import TableParametersManager
from datalakebundle.table.name.TableNameTemplateGetter import TableNameTemplateGetter


@input_decorator_function
def read_table(identifier: str):
    def wrapper(container: ContainerInterface):
        table_name_template_getter: TableNameTemplateGetter = container.get(TableNameTemplateGetter)
        table_parameters_manager: TableParametersManager = container.get(TableParametersManager)
        table_name_template = table_name_template_getter.get_template_for_read()
        table_parameters = table_parameters_manager.get_or_parse(table_name_template, identifier)

        logger: Logger = container.get("datalakebundle.logger")
        logger.info(f"Reading table `{table_parameters.full_table_name}`")

        spark: SparkSession = container.get(SparkSession)

        return spark.read.table(table_parameters.full_table_name)

    return wrapper



from logging import Logger
from pyspark.sql.dataframe import DataFrame
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.schema.SchemaChecker import SchemaChecker
from datalakebundle.table.write.DataWriter import DataWriter
from datalakebundle.table.create.TableCreator import TableCreator


class TableAppender:
    def __init__(self, logger: Logger, schema_checker: SchemaChecker, table_creator: TableCreator, data_writer: DataWriter):
        self.__logger = logger
        self.__schema_checker = schema_checker
        self.__table_creator = table_creator
        self.__data_writer = data_writer

    def append(self, result: DataFrame, table_definition: TableDefinition, options: dict):
        output_table_name = table_definition.full_table_name

        self.__schema_checker.check_with_metadata(result, table_definition)

        self.__logger.info(f"Data to be appended into table: {output_table_name}")

        self.__table_creator.create_if_not_exists(table_definition)

        self.__logger.info(f"Appending data to table: {output_table_name}")

        self.__data_writer.append(result, output_table_name, table_definition.schema, options)

        self.__logger.info(f"Data successfully appended to: {output_table_name}")



from pyspark.sql import types as t
from typing import Union


class TableSchemaGenerator:
    def generate(self, schema: t.StructType) -> str:
        indent = "    "

        def generate_schema_recursively(element: Union[t.StructType, t.ArrayType, t.StructField], last_struct_type_indent: str = indent):
            schema_string = ""

            if isinstance(element, t.StructType):
                schema_string += last_struct_type_indent + "t.StructType(\n"
                schema_string += last_struct_type_indent + indent + "[\n"

                for field in element:
                    schema_string += generate_schema_recursively(field, last_struct_type_indent)

                schema_string += last_struct_type_indent + indent + "],\n"
                schema_string += last_struct_type_indent + "),\n"

            elif isinstance(element, t.ArrayType):
                if isinstance(element.elementType, (t.StructType, t.ArrayType)):
                    schema_string += last_struct_type_indent + "t.ArrayType(\n"
                    schema_string += generate_schema_recursively(element.elementType, last_struct_type_indent + indent)  # pyre-ignore[6]
                    schema_string += last_struct_type_indent + "),\n"

                else:
                    schema_string += last_struct_type_indent + f"t.ArrayType(t.{element.elementType}),\n"

            elif isinstance(element, t.StructField):
                if isinstance(element.dataType, t.StructType):
                    schema_string += last_struct_type_indent + 2 * indent + "t.StructField(\n"
                    schema_string += last_struct_type_indent + 3 * indent + f'"{element.name}",\n'
                    schema_string += generate_schema_recursively(element.dataType, last_struct_type_indent + 3 * indent)  # pyre-ignore[6]
                    schema_string += last_struct_type_indent + 2 * indent + "),\n"

                elif isinstance(element.dataType, t.ArrayType):
                    if isinstance(element.dataType.elementType, (t.StructType, t.ArrayType)):
                        schema_string += last_struct_type_indent + 2 * indent + "t.StructField(\n"
                        schema_string += last_struct_type_indent + 3 * indent + f'"{element.name}",\n'
                        schema_string += generate_schema_recursively(
                            element.dataType, last_struct_type_indent + 3 * indent  # pyre-ignore[6]
                        )
                        schema_string += last_struct_type_indent + 2 * indent + "),\n"

                    else:
                        schema_string += (
                            last_struct_type_indent
                            + 2 * indent
                            + f't.StructField("{element.name}", t.ArrayType(t.{element.dataType.elementType})),\n'  # pyre-ignore[16]
                        )

                else:
                    schema_string += last_struct_type_indent + 2 * indent + f't.StructField("{element.name}", t.{element.dataType}),\n'

            return schema_string

        def remove_top_level_struct_type(schema_string):
            return "\n".join(list(schema_string.split("\n")[1:-2])) + "\n"

        table_schema = ""

        table_schema += "def get_schema():\n"
        table_schema += "    return dp.TableSchema(\n"

        schema_string = generate_schema_recursively(schema)
        schema_string = remove_top_level_struct_type(schema_string)
        table_schema += schema_string

        table_schema += '        # primary_key="", # INSERT PRIMARY KEY(s) HERE (OPTIONAL)\n'
        table_schema += '        # partition_by="" # INSERT PARTITION KEY(s) HERE (OPTIONAL)\n'
        table_schema += "        # tbl_properties={} # INSERT TBLPROPERTIES HERE (OPTIONAL)\n"
        table_schema += "    )\n"

        return table_schema



import json


class TableParameters:
    def __init__(
        self,
        db_identifier: str,
        db_name: str,
        table_identifier: str,
        table_name: str,
        target_path: str,
        **kwargs,
    ):
        self.__db_identifier = db_identifier
        self.__db_name = db_name
        self.__table_identifier = table_identifier
        self.__table_name = table_name
        self.__target_path = target_path
        self.__custom_fields = kwargs

    @property
    def identifier(self):
        return self.__db_identifier + "." + self.__table_identifier

    @property
    def db_identifier(self):
        return self.__db_identifier

    @property
    def db_name(self):
        return self.__db_name

    @property
    def table_identifier(self):
        return self.__table_identifier

    @property
    def table_name(self):
        return self.__table_name

    @property
    def target_path(self):
        return self.__target_path

    @property
    def full_table_name(self):
        return self.__db_name + "." + self.__table_name

    def get_custom_fields(self):
        return self.__custom_fields

    def to_dict(self):
        return {
            **self.__custom_fields,
            **{
                "identifier": self.identifier,
                "db_identifier": self.__db_identifier,
                "db_name": self.__db_name,
                "table_identifier": self.__table_identifier,
                "table_name": self.__table_name,
                "target_path": self.__target_path,
            },
        }

    def __getattr__(self, item):
        if item not in self.__custom_fields:
            raise Exception(f'Unexpected attribute: "{item}"')

        return self.__custom_fields[item]

    def __eq__(self, other: "TableParameters"):
        return self.to_dict() == other.to_dict()

    def __repr__(self):
        return json.dumps(self.to_dict(), indent=4)



# pylint: disable = super-init-not-called
from logging import Logger
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from daipecore.decorator.OutputDecorator import OutputDecorator
from injecta.container.ContainerInterface import ContainerInterface


class TestingStorage:
    result = None


@DecoratedDecorator
class dummy_saver(OutputDecorator):  # pylint: disable = invalid-name
    def __init__(self, identifier: str):
        self._identifier = identifier

    def process_result(self, result, container: ContainerInterface):
        logger: Logger = container.get("datalakebundle.logger")

        logger.info(f"Saving into {self._identifier}")

        TestingStorage.result = result



# pylint: disable = super-init-not-called
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from pyspark.sql import DataFrame
import pandas as pd
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator
from datalakebundle.notebook.decorator.DuplicateColumnsChecker import DuplicateColumnsChecker
from injecta.container.ContainerInterface import ContainerInterface
from typing import Union


@DecoratedDecorator
class transformation(DataFrameReturningDecorator):  # pylint: disable = invalid-name
    def __init__(self, *args, display=False, check_duplicate_columns=True):
        self._args = args
        self._display = display
        self._check_duplicate_columns = check_duplicate_columns

    def after_execution(self, container: ContainerInterface):
        if self._result is None:
            return

        if isinstance(self.result, (DataFrame, pd.DataFrame)):
            self.__process_df(container)
        else:
            raise TypeError(
                f"Result is of type: '{type(self._result).__name__}', only Spark and Pandas DataFrames are valid results in a @transformation."
            )

    def __process_df(self, container: ContainerInterface):
        if self._check_duplicate_columns and container.get_parameters().datalakebundle.notebook.duplicate_columns_check.enabled is True:
            duplicate_columns_checker: DuplicateColumnsChecker = container.get(DuplicateColumnsChecker)
            data_frame_decorators = filter(lambda decorator_arg: isinstance(decorator_arg, DataFrameReturningDecorator), self._args)
            duplicate_columns_checker.check(self._result, data_frame_decorators)

        if self._display and container.get_parameters().datalakebundle.notebook.display.enabled is True:
            self.__display(container, self.__get_display_service(self._result))

        self._set_global_dataframe()

    def __display(self, container: ContainerInterface, show_method: str):
        show_service = container.get(show_method)
        show_service.show(self._result)

    def __get_display_service(self, df: Union[DataFrame, pd.DataFrame]) -> str:
        if isinstance(df, DataFrame):
            return "pysparkbundle.dataframe.show_method"
        return "daipecore.pandas.dataframe.show_method"



from typing import Optional
from datalakebundle.table.parameters import primitive_value
from datalakebundle.table.parameters.FieldsResolver import FieldsResolver
from datalakebundle.table.parameters.TableParameters import TableParameters
from datalakebundle.table.name.TableNamesPreparer import TableNamesPreparer


class TableParametersParser:

    __base_fields = ["identifier", "db_identifier", "db_name", "table_identifier", "table_name", "target_path"]

    def __init__(self, table_names_preparer: TableNamesPreparer):
        self.__fields_resolver = FieldsResolver()
        self.__table_names_preparer = table_names_preparer

    def parse(self, table_name_template: str, identifier: str, defaults: Optional[dict] = None, explicit_parameters: Optional[dict] = None):
        defaults = defaults or {}
        explicit_parameters = explicit_parameters or {}
        table_names = self.__table_names_preparer.prepare(table_name_template, identifier)

        all_fields = {**table_names.to_dict(), **explicit_parameters}

        for name, val in self.__filter_primitive(all_fields, defaults).items():
            all_fields[name] = primitive_value.evaluate(val, all_fields)

        all_fields = self.__fields_resolver.resolve(all_fields, defaults)

        return TableParameters(
            table_names.db_identifier,
            all_fields["db_name"],
            table_names.table_identifier,
            all_fields["table_name"],
            all_fields["target_path"],
            **{k: v for k, v in all_fields.items() if k not in self.__base_fields}
        )

    def __filter_primitive(self, all_fields: dict, defaults: dict):
        return {name: val for name, val in defaults.items() if not isinstance(val, dict) and name not in all_fields}



import sys
from argparse import ArgumentParser, Namespace
from logging import Logger
from time import sleep
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.UnknownTableException import UnknownTableException
from datalakebundle.table.parameters.TableParametersManager import TableParametersManager
from datalakebundle.table.name.TableNameTemplateGetter import TableNameTemplateGetter
from datalakebundle.table.delete.TableDeleter import TableDeleter


class TableDeleterCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_name_template_getter: TableNameTemplateGetter,
        table_parameters_manager: TableParametersManager,
        table_existence_checker: TableExistenceChecker,
        table_deleter: TableDeleter,
    ):
        self.__logger = logger
        self.__table_name_template_getter = table_name_template_getter
        self.__table_parameters_manager = table_parameters_manager
        self.__table_existence_checker = table_existence_checker
        self.__table_deleter = table_deleter

    def get_command(self) -> str:
        return "datalake:table:delete-including-data"

    def get_description(self):
        return "Deletes a Hive table including all data"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")
        argument_parser.add_argument(
            "-s",
            "--skip-countdown",
            action="store_true",
            help="Skip data deletion countdown",
        )

    def run(self, input_args: Namespace):
        table_name_template = self.__table_name_template_getter.get_template_for_write()
        table_parameters = self.__table_parameters_manager.get_or_parse(table_name_template, input_args.identifier)

        if not self.__table_existence_checker.table_exists(table_parameters.db_name, table_parameters.table_name):
            self.__logger.error(f"Hive table {table_parameters.full_table_name} does NOT exists")
            sys.exit(1)

        if input_args.skip_countdown is False:
            self.__logger.info("Use the --skip-countdown switch to delete existing data immediately")

            countdown = 10

            for i in range(countdown):
                self.__logger.warning(f"Table data will be deleted in {countdown - i}s. Use CTRL+C to cancel.")
                sleep(1)

        try:
            self.__table_deleter.delete_including_data(table_parameters)
        except UnknownTableException as e:
            self.__logger.error(str(e))
            sys.exit(1)



import pyspark.sql.types as t

expected_schema_array2 = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(t.ArrayType(t.StringType())),
        ),
    ],
)



import yaml
from simpleeval import simple_eval


class Expression:
    yaml_tag = "!expr"
    yaml_loader = yaml.SafeLoader

    def __init__(self, val):
        self.val = val

    @classmethod
    def from_yaml(cls, loader, node):  # pylint: disable = unused-argument
        return cls(node.value)

    def evaluate(self, variables: dict):
        return simple_eval(self.val, names=variables)



from pyspark.sql.session import SparkSession


class TableExistenceChecker:
    def __init__(self, spark: SparkSession):
        self.__spark = spark

    def table_exists(self, db_name: str, table_name: str) -> bool:
        return self.__spark.sql(f'SHOW TABLES IN {db_name} LIKE "{table_name}"').collect() != []



import re


def fill_template(template: str, replacements: dict):
    def convert(matches):
        placeholder = matches.group(1)

        if placeholder not in replacements:
            raise Exception("Value for placeholder {" + placeholder + "} not defined")

        return str(replacements[placeholder])

    return re.sub(r"\{([^\}]+)\}", convert, template)



import json
from logging import Logger
from typing import List
from pyspark.sql.session import SparkSession
from pyspark.sql.functions import col

from datalakebundle.table.create.TableDefinition import TableDefinition
from deepdiff import DeepDiff


class MetadataChecker:
    def __init__(self, default_properties: List[str], logger: Logger, spark: SparkSession):
        self.__default_properties = default_properties
        self.__logger = logger
        self.__spark = spark

    def check(self, table_definition: TableDefinition):
        tables = self.__spark.sql(f"SHOW TABLES IN {table_definition.db_name} LIKE '{table_definition.table_name}'")

        if not tables.rdd.isEmpty():
            self.__check_partition_by(table_definition)
            self.__check_tbl_properties(table_definition)

    def __check_partition_by(self, table_definition: TableDefinition):
        partitions_df = self.__spark.sql(f"DESCRIBE TABLE {table_definition.full_table_name}")
        partitions_df = partitions_df.filter(col("col_name").contains("Part "))

        if not partitions_df.rdd.isEmpty():
            partitions_col = partitions_df.select("data_type").collect()
            partitions = [row.data_type for row in partitions_col]

            ddiff = DeepDiff(partitions, table_definition.partition_by)

            if ddiff:
                unexpected_keys: dict = ddiff.get("iterable_item_added")
                missing_keys: dict = ddiff.get("iterable_item_removed")
                values_changed: dict = ddiff.get("values_changed")

                extra = {}

                if unexpected_keys:
                    extra["unexpected_keys"] = list(unexpected_keys.values())

                if missing_keys:
                    extra["missing_keys"] = list(missing_keys.values())

                if values_changed:
                    extra["values_changed"] = [f'{v["old_value"]} changed to {v["new_value"]}' for v in values_changed.values()]

                self.__create_warning(
                    "Table partitioning does NOT match schema",
                    extra,
                    table_definition,
                )

    def __check_tbl_properties(self, table_definition: TableDefinition):
        properties_df = self.__spark.sql(f"SHOW TBLPROPERTIES {table_definition.full_table_name}")
        tbl_properties = dict(properties_df.rdd.map(lambda x: (x.key, x.value)).collect())

        for prop in self.__default_properties:
            tbl_properties.pop(prop, None)

        ddiff = DeepDiff(tbl_properties, table_definition.tbl_properties)

        if ddiff:
            unexpected_keys: list = ddiff.get("dictionary_item_added")
            missing_keys: list = ddiff.get("dictionary_item_removed")
            values_changed: dict = ddiff.get("values_changed")

            extra = {}

            def remove_wrapper(key):
                return key[5:-1]

            if missing_keys:
                extra["missing_keys"] = [remove_wrapper(key) for key in missing_keys]

            if unexpected_keys:
                extra["unexpected_keys"] = [remove_wrapper(key) for key in unexpected_keys]

            if values_changed:
                extra["values_changed"] = {
                    remove_wrapper(key): value.get("new_value", "Not found") for key, value in values_changed.items()
                }

            self.__create_warning(
                "TBL_PROPERTIES do NOT match table schema",
                extra,
                table_definition,
            )

    def __create_warning(self, warning_message: str, extra: dict, table_definition: TableDefinition):
        extra["table"] = table_definition.full_table_name

        report = json.dumps(extra, indent=4).replace('"', "")

        self.__logger.warning(f"{warning_message}\n" f"{report}")



# pylint: disable = invalid-name
import pyspark.sql.types as t
from datalakebundle.table.schema.TableSchemaGenerator import TableSchemaGenerator

schema = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField("FIELD3", t.DoubleType()),
        t.StructField(
            "STRUCT1",
            t.StructType(
                [
                    t.StructField("NESTED_FIELD1", t.StringType()),
                    t.StructField(
                        "STRUCT2",
                        t.StructType(
                            [
                                t.StructField("NESTED_FIELD2", t.StringType()),
                            ],
                        ),
                    ),
                ],
            ),
        ),
    ],
)

expected_result = """def get_schema():
    return dp.TableSchema(
        [
            t.StructField("FIELD1", t.IntegerType()),
            t.StructField("FIELD2", t.DoubleType()),
            t.StructField("FIELD3", t.DoubleType()),
            t.StructField(
                "STRUCT1",
                t.StructType(
                    [
                        t.StructField("NESTED_FIELD1", t.StringType()),
                        t.StructField(
                            "STRUCT2",
                            t.StructType(
                                [
                                    t.StructField("NESTED_FIELD2", t.StringType()),
                                ],
                            ),
                        ),
                    ],
                ),
            ),
        ],
        # primary_key="", # INSERT PRIMARY KEY(s) HERE (OPTIONAL)
        # partition_by="" # INSERT PARTITION KEY(s) HERE (OPTIONAL)
        # tbl_properties={} # INSERT TBLPROPERTIES HERE (OPTIONAL)
    )
"""

assert TableSchemaGenerator().generate(schema) == expected_result

schema = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField("FIELD3", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(
                t.StructType(
                    [
                        t.StructField("NESTED_ARRAY_FIELD1", t.StringType()),
                        t.StructField("NESTED_ARRAY_FIELD2", t.StringType()),
                        t.StructField("NESTED_ARRAY_FIELD3", t.ArrayType(t.StringType())),
                    ],
                ),
            ),
        ),
    ],
)

expected_result = """def get_schema():
    return dp.TableSchema(
        [
            t.StructField("FIELD1", t.IntegerType()),
            t.StructField("FIELD2", t.DoubleType()),
            t.StructField("FIELD3", t.DoubleType()),
            t.StructField(
                "ARRAY1",
                t.ArrayType(
                    t.StructType(
                        [
                            t.StructField("NESTED_ARRAY_FIELD1", t.StringType()),
                            t.StructField("NESTED_ARRAY_FIELD2", t.StringType()),
                            t.StructField("NESTED_ARRAY_FIELD3", t.ArrayType(t.StringType())),
                        ],
                    ),
                ),
            ),
        ],
        # primary_key="", # INSERT PRIMARY KEY(s) HERE (OPTIONAL)
        # partition_by="" # INSERT PARTITION KEY(s) HERE (OPTIONAL)
        # tbl_properties={} # INSERT TBLPROPERTIES HERE (OPTIONAL)
    )
"""

assert TableSchemaGenerator().generate(schema) == expected_result

schema = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField("FIELD3", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(
                t.ArrayType(t.StringType()),
            ),
        ),
    ],
)

expected_result = """def get_schema():
    return dp.TableSchema(
        [
            t.StructField("FIELD1", t.IntegerType()),
            t.StructField("FIELD2", t.DoubleType()),
            t.StructField("FIELD3", t.DoubleType()),
            t.StructField(
                "ARRAY1",
                t.ArrayType(
                    t.ArrayType(t.StringType()),
                ),
            ),
        ],
        # primary_key="", # INSERT PRIMARY KEY(s) HERE (OPTIONAL)
        # partition_by="" # INSERT PARTITION KEY(s) HERE (OPTIONAL)
        # tbl_properties={} # INSERT TBLPROPERTIES HERE (OPTIONAL)
    )
"""

assert TableSchemaGenerator().generate(schema) == expected_result

schema = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField("FIELD3", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(
                t.ArrayType(
                    t.StructType(
                        [
                            t.StructField(
                                "VERY_BADLY_NESTED_ARRAY_OF_ARRAY_OF_ARRAY_OF_DOUBLES",
                                t.ArrayType(
                                    t.ArrayType(
                                        t.ArrayType(t.DoubleType()),
                                    ),
                                ),
                            ),
                        ],
                    ),
                ),
            ),
        ),
    ],
)

expected_result = """def get_schema():
    return dp.TableSchema(
        [
            t.StructField("FIELD1", t.IntegerType()),
            t.StructField("FIELD2", t.DoubleType()),
            t.StructField("FIELD3", t.DoubleType()),
            t.StructField(
                "ARRAY1",
                t.ArrayType(
                    t.ArrayType(
                        t.StructType(
                            [
                                t.StructField(
                                    "VERY_BADLY_NESTED_ARRAY_OF_ARRAY_OF_ARRAY_OF_DOUBLES",
                                    t.ArrayType(
                                        t.ArrayType(
                                            t.ArrayType(t.DoubleType()),
                                        ),
                                    ),
                                ),
                            ],
                        ),
                    ),
                ),
            ),
        ],
        # primary_key="", # INSERT PRIMARY KEY(s) HERE (OPTIONAL)
        # partition_by="" # INSERT PARTITION KEY(s) HERE (OPTIONAL)
        # tbl_properties={} # INSERT TBLPROPERTIES HERE (OPTIONAL)
    )
"""

assert TableSchemaGenerator().generate(schema) == expected_result



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['datalakebundle',
 'datalakebundle.delta',
 'datalakebundle.notebook.decorator',
 'datalakebundle.notebook.decorator.tests',
 'datalakebundle.notebook.lineage',
 'datalakebundle.notebook.lineage.argument',
 'datalakebundle.table',
 'datalakebundle.table.class_',
 'datalakebundle.table.create',
 'datalakebundle.table.delete',
 'datalakebundle.table.identifier',
 'datalakebundle.table.name',
 'datalakebundle.table.optimize',
 'datalakebundle.table.parameters',
 'datalakebundle.table.read',
 'datalakebundle.table.schema',
 'datalakebundle.table.schema.DiffGeneratorTest_stubs',
 'datalakebundle.table.upsert',
 'datalakebundle.table.write',
 'datalakebundle.table.write.tests',
 'datalakebundle.test']

package_data = \
{'': ['*'], 'datalakebundle': ['_config/*']}

install_requires = \
['console-bundle>=0.5,<0.6',
 'daipe-core>=1.2,<2.0',
 'deepdiff>=5,<6',
 'injecta>=0.10.0,<0.11.0',
 'pyfony-bundles>=0.4.0,<0.5.0',
 'pyspark-bundle>=1.2,<2.0',
 'simpleeval>=0.9.10,<1.0.0']

entry_points = \
{'pyfony.bundle': ['create = datalakebundle.DataLakeBundle:DataLakeBundle']}

setup_kwargs = {
    'name': 'datalake-bundle',
    'version': '1.3.0.dev1',
    'description': 'DataLake tables management bundle for the Daipe Framework',
    'long_description': '# Datalake bundle\n\nThis bundle allows you to easily create and manage datalake(house) based on the [Daipe Framework](https://www.daipe.ai).  \n\n## Resources\n\n* [Documentation](https://docs.daipe.ai/data-pipelines-workflow/managing-datalake/)\n',
    'author': 'Jiri Koutny',
    'author_email': 'jiri.koutny@datasentics.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/daipe-ai/datalake-bundle',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pyspark.sql.types import StructType


class TableDefinition:
    def __init__(
        self,
        db_name: str,
        table_name: str,
        schema: StructType,
        primary_key: list,
        partition_by: list,
        target_path: str,
        tbl_properties: dict,
    ):
        if not isinstance(primary_key, list):
            raise Exception(f"Invalid primary key: {primary_key}")

        if not isinstance(partition_by, list):
            raise Exception(f"Invalid partition_by: {partition_by}")

        if not isinstance(tbl_properties, dict):
            raise Exception(f"Invalid tbl_properties: {tbl_properties}")

        self.__db_name = db_name
        self.__table_name = table_name
        self.__schema = schema
        self.__primary_key = primary_key
        self.__partition_by = partition_by
        self.__target_path = target_path
        self.__tbl_properties = tbl_properties

    @property
    def db_name(self):
        return self.__db_name

    @property
    def table_name(self):
        return self.__table_name

    @property
    def full_table_name(self):
        return self.__db_name + "." + self.__table_name

    @property
    def schema(self) -> StructType:
        return self.__schema

    @property
    def primary_key(self) -> list:
        return self.__primary_key

    @property
    def partition_by(self) -> list:
        return self.__partition_by

    @property
    def target_path(self):
        return self.__target_path

    @property
    def tbl_properties(self) -> dict:
        return self.__tbl_properties



class TableNameTemplateGetter:
    def __init__(self, table_name_read_template: str, table_name_write_template: str):
        self.__table_name_read_template = table_name_read_template
        self.__table_name_write_template = table_name_write_template

    def get_template_for_read(self) -> str:
        return self.__table_name_read_template

    def get_template_for_write(self) -> str:
        return self.__table_name_write_template



from daipecore.decorator.InputDecorator import InputDecorator


class DataFrameReturningDecorator(InputDecorator):
    def _set_global_dataframe(self):
        self._function.__globals__[self._function.__name__ + "_df"] = self._result



from typing import Optional
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType
from pysparkbundle.filesystem.FilesystemInterface import FilesystemInterface
from datalakebundle.table.create.TableDefinition import TableDefinition


class DeltaStorage:
    def __init__(
        self,
        filesystem: FilesystemInterface,
        spark: SparkSession,
    ):
        self.__filesystem = filesystem
        self.__spark = spark

    def create_table(self, table_definition: TableDefinition, options: Optional[dict] = None):
        df = self.__create_empty_df(table_definition.schema)

        self.__create(df, table_definition, "errorifexists", options or {})

    def recreate_table(self, table_definition: TableDefinition, options: Optional[dict] = None):
        self.__spark.sql(f"DROP TABLE IF EXISTS {table_definition.full_table_name}")
        self.__filesystem.delete(table_definition.target_path, True)
        self.create_table(table_definition, options or {})

    def overwrite_data(self, df: DataFrame, full_table_name: str, partition_by: list, options: dict):
        options = options or {}
        base_options = {"overwriteSchema": "true"}
        merged_options = {**options, **base_options}

        self.__save_as_table(df, full_table_name, "overwrite", partition_by, merged_options)

    def __create(self, df: DataFrame, table_definition: TableDefinition, mode: str, options: dict):
        base_options = {"path": table_definition.target_path}
        merged_options = {**options, **base_options}

        self.__save_as_table(df, table_definition.full_table_name, mode, table_definition.partition_by, merged_options)

    def __save_as_table(self, df: DataFrame, full_table_name: str, mode: str, partition_by: list, options: dict):
        (df.write.partitionBy(partition_by).format("delta").options(**options).mode(mode).saveAsTable(full_table_name))

    def __create_empty_df(self, schema: StructType):
        return self.__spark.createDataFrame([], schema)



from box import Box
from datalakebundle.table.parameters.ValueResolverFactory import ValueResolverFactory


class FieldsResolver:
    def __init__(self):
        self.__value_resolver_factory = ValueResolverFactory()

    def resolve(self, all_fields: dict, defaults: dict):
        resolver_keys = [name for name, val in defaults.items() if isinstance(val, dict) and name not in all_fields]

        unsuccessfully_assigned_fields = set()

        while resolver_keys:
            name = resolver_keys.pop(0)
            resolver = self.__value_resolver_factory.create(defaults[name])
            dependent_fields = resolver.get_depending_fields()

            if dependent_fields - set(all_fields.keys()) == set():
                unsuccessfully_assigned_fields = set()
                all_fields[name] = resolver.resolve(Box(all_fields))
            else:
                if name in unsuccessfully_assigned_fields:
                    raise Exception(
                        f'Infinite assignment loop detected. Check get_depending_fields() of {defaults[name]["resolver_class"]}'
                    )

                unsuccessfully_assigned_fields.add(name)
                resolver_keys.append(name)

        return all_fields



import pyspark.sql.types as t

df_schema_array3 = t.StructType(
    [
        t.StructField("FIELD1", t.IntegerType()),
        t.StructField("FIELD2", t.DoubleType()),
        t.StructField(
            "ARRAY1",
            t.ArrayType(
                t.StructType(
                    [
                        t.StructField("FIELD4", t.IntegerType()),
                        t.StructField("FIELD5", t.IntegerType()),
                    ]
                )
            ),
        ),
    ],
)



import os
from daipecore.decorator.notebook_function import notebook_function
from datalakebundle.table.write.tests.dummy_saver import dummy_saver, TestingStorage

os.environ["APP_ENV"] = "test"


@notebook_function()
@dummy_saver("my_sample_table")
def load_data():
    return 155


@notebook_function(load_data)
def process(number: int):
    assert number == 155
    assert TestingStorage.result == 155



from typing import Union, Optional
from pyspark.sql.types import StructType


def _get_list_value(value, invalid_message: str):
    if value is None:
        return []

    if isinstance(value, str):
        return [value]

    if not isinstance(value, str) and not isinstance(value, list):
        raise Exception(invalid_message.format(value=value))

    return value


def _get_tbl_properties(tbl_properties):
    if tbl_properties is None:
        tbl_properties = tbl_properties or {}
    elif not isinstance(tbl_properties, dict):
        raise Exception(f"Invalid tbl_properties: {tbl_properties}")

    for k, v in tbl_properties.items():
        if not isinstance(k, str) or not isinstance(v, str):
            raise Exception(f"Invalid tbl_properties - keys and values not strings: {tbl_properties}")

    return tbl_properties


class TableSchema(StructType):
    def __init__(
        self,
        fields: list,
        primary_key: Optional[Union[str, list]] = None,
        partition_by: Optional[Union[str, list]] = None,
        tbl_properties: Optional[dict] = None,
    ):
        super().__init__(fields)

        self.__primary_key = _get_list_value(primary_key, "Invalid primary key: {value}")
        self.__partition_by = _get_list_value(partition_by, "Invalid partition by: {value}")
        self.__tbl_properties = _get_tbl_properties(tbl_properties)

    @property
    def primary_key(self) -> list:
        return self.__primary_key

    @property
    def partition_by(self) -> list:
        return self.__partition_by

    @property
    def tbl_properties(self) -> dict:
        return self.__tbl_properties

    @classmethod
    def typeName(cls):
        return "struct"



import json
from logging import Logger
from datalakebundle.table.schema.DiffGenerator import DiffGenerator
from pyspark.sql.dataframe import DataFrame
from datalakebundle.table.create.TableDefinition import TableDefinition
from datalakebundle.table.schema.MetadataChecker import MetadataChecker


class SchemaChecker:
    def __init__(self, logger: Logger, metadata_checker: MetadataChecker, diff_generator: DiffGenerator):
        self.__logger = logger
        self.__metadata_checker = metadata_checker
        self.__diff_generator = diff_generator

    def check_with_metadata(self, df: DataFrame, table_definition: TableDefinition):
        self.check(df, table_definition)
        self.check_metadata(table_definition)

    def check(self, df: DataFrame, table_definition: TableDefinition):
        extra = {"table": table_definition.full_table_name, "diff": self.__diff_generator.generate(df.schema, table_definition.schema)}

        if extra["diff"]:
            error_message = "Table and dataframe schemas do NOT match"
            report = json.dumps(extra, indent=4).replace('"', "")

            self.__logger.error(f"{error_message}\n{report}")
            raise Exception(error_message)

    def check_metadata(self, table_definition: TableDefinition):
        self.__metadata_checker.check(table_definition)



# pylint: disable = super-init-not-called
from daipecore.decorator.DecoratedDecorator import DecoratedDecorator
from daipecore.decorator.OutputDecorator import OutputDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import DataFrame
from datalakebundle.table.create.TableDefinitionFactory import TableDefinitionFactory
from datalakebundle.table.schema.TableSchema import TableSchema
from datalakebundle.table.write.TableUpserter import TableUpserter


@DecoratedDecorator
class table_upsert(OutputDecorator):  # pylint: disable = invalid-name
    def __init__(self, identifier: str, table_schema: TableSchema):
        self.__identifier = identifier
        self.__table_schema = table_schema

    def process_result(self, result: DataFrame, container: ContainerInterface):
        table_definition_factory: TableDefinitionFactory = container.get(TableDefinitionFactory)
        table_upserter: TableUpserter = container.get(TableUpserter)

        if not isinstance(self.__table_schema, TableSchema):
            raise Exception(f"Invalid table schema: {self.__table_schema}")

        if not self.__table_schema.primary_key:
            raise Exception("Table schema must have primary key defined for upsert to work properly")

        table_definition = table_definition_factory.create_from_table_schema(self.__identifier, self.__table_schema)

        table_upserter.upsert(result, table_definition)



import collections
import pandas as pd
from logging import Logger
from typing import Iterable, Union
from pyspark.sql.dataframe import DataFrame
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator


class DuplicateColumnsChecker:
    def __init__(
        self,
        logger: Logger,
    ):
        self.__logger = logger

    def check(self, df: Union[DataFrame, pd.DataFrame], result_decorators: Iterable[DataFrameReturningDecorator]):
        duplicate_fields = [field_name for field_name, count in collections.Counter(self.__get_fields(df)).items() if count > 1]

        if not duplicate_fields:
            return

        fields2_tables = {}

        for result_decorator in result_decorators:
            source_df = result_decorator.result
            for field_name in self.__get_fields(source_df):
                if field_name not in fields2_tables:
                    fields2_tables[field_name] = []

                fields2_tables[field_name].append(result_decorator.function.__name__)

        for duplicate_field in duplicate_fields:
            self.__logger.error(f"Duplicate field '{duplicate_field}'", extra={"source_functions": fields2_tables[duplicate_field]})

        fields_string = ", ".join(duplicate_fields)
        raise Exception(f"Duplicate output column(s): {fields_string}. Disable by setting @transformation(check_duplicate_columns=False)")

    def __get_fields(self, df: Union[DataFrame, pd.DataFrame]) -> Iterable[str]:
        if isinstance(df, DataFrame):
            return map(lambda field: field.name.lower(), df.schema.fields)

        return map(lambda field: field.lower(), df.columns)



from box import Box
from datalakebundle.table.identifier.ValueResolverInterface import ValueResolverInterface


class SimpleTargetPathResolver(ValueResolverInterface):
    def __init__(self, base_path: str):
        self.__base_path = base_path

    def resolve(self, raw_table_parameters: Box):
        return self.__base_path + "/" + raw_table_parameters.db_identifier + "/" + raw_table_parameters.table_identifier + ".delta"

    def get_depending_fields(self) -> set:
        return {"db_identifier", "table_identifier"}



from abc import ABC, abstractmethod
from box import Box


class ValueResolverInterface(ABC):
    @abstractmethod
    def resolve(self, raw_table_parameters: Box):
        pass

    def get_depending_fields(self) -> set:
        return set()



# pylint: disable = import-outside-toplevel
import unittest
from datalakebundle.table.schema.DiffGenerator import DiffGenerator


class DiffGeneratorTest(unittest.TestCase):
    def setUp(self):
        self.__diff_generator = DiffGenerator()

    def test_no_diff(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema import df_schema
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array2 import df_schema_array2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array_reordered_changed_case import (
            df_schema_array_reordered_changed_case,
        )

        self.assertListEqual(self.__diff_generator.generate(df_schema, df_schema), [])
        self.assertListEqual(self.__diff_generator.generate(df_schema_array2, df_schema_array2), [])
        self.assertListEqual(self.__diff_generator.generate(df_schema_array2, df_schema_array_reordered_changed_case), [])

    def test_changed_attributes(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema import df_schema
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema import expected_schema
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema2 import expected_schema2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema3 import expected_schema3

        self.assertListEqual(self.__diff_generator.generate(df_schema, expected_schema), ["FIELD2['type'] changed from DATE to STRING"])

        self.assertListEqual(
            self.__diff_generator.generate(df_schema, expected_schema2), ["STRUCT1.NESTED_FIELD1['type'] changed from DATE to STRING"]
        )

        self.assertListEqual(
            self.__diff_generator.generate(df_schema, expected_schema3),
            ["STRUCT1.STRUCT2.NESTED_FIELD2['type'] changed from DATE to STRING"],
        )

    def test_added_columns_to_df_schema(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema2 import df_schema2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema4 import expected_schema4
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema3 import df_schema3
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema5 import expected_schema5

        self.assertListEqual(self.__diff_generator.generate(df_schema2, expected_schema4), ["root unexpected field: FIELD3"])

        self.assertListEqual(
            self.__diff_generator.generate(df_schema3, expected_schema4),
            [
                "STRUCT1['name'] changed from STRUCT1 to FIELD3",
                "STRUCT1['type'] changed from array to string",
                "root unexpected field: STRUCT1",
            ],
        )

        self.assertListEqual(
            self.__diff_generator.generate(df_schema3, expected_schema5),
            [
                "STRUCT1 unexpected field: STRUCT3",
                "STRUCT1.STRUCT2 missing field: NESTED_FIELD0",
                "STRUCT1.STRUCT2 missing field: NESTED_FIELD00",
            ],
        )

    def test_completely_different_schemas(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema4 import df_schema4
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema6 import expected_schema6

        self.assertListEqual(
            self.__diff_generator.generate(df_schema4, expected_schema6),
            [
                "FIELD6['name'] changed from FIELD6 to FIELD1",
                "FIELD6['type'] changed from struct (STRUCT1) to integer",
            ],
        )

    def test_schemas_with_arrays(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array import df_schema_array
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema_array import expected_schema_array
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array2 import df_schema_array2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array3 import df_schema_array3
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema_array2 import expected_schema_array2

        self.assertListEqual(
            self.__diff_generator.generate(df_schema_array, expected_schema_array),
            ["ARRAY1['elementType'] changed from array to string"],
        )
        self.assertListEqual(
            self.__diff_generator.generate(df_schema_array2, expected_schema_array),
            ["ARRAY1.array['elementType'] changed from string to array"],
        )
        self.assertListEqual(
            self.__diff_generator.generate(df_schema_array3, expected_schema_array2),
            [
                "ARRAY1.array['type'] changed from array to struct",
                "Unexpected field ARRAY1.array.struct",
                "Missing field ARRAY1.array.array",
            ],
        )

    def test_ignore_nullable(self):
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema import df_schema
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_nullable import df_schema_nullable
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array2 import df_schema_array2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_array_nullable2 import df_schema_array_nullable2
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.df_schema_nullable3 import df_schema_nullable3
        from datalakebundle.table.schema.DiffGeneratorTest_stubs.expected_schema_nullable3 import expected_schema_nullable3

        self.assertListEqual(self.__diff_generator.generate(df_schema, df_schema_nullable), [])
        self.assertListEqual(self.__diff_generator.generate(df_schema_array2, df_schema_array_nullable2), [])
        self.assertListEqual(
            self.__diff_generator.generate(df_schema_nullable3, expected_schema_nullable3),
            ["STRUCT1.['nullable']['name'] changed from ['NULLABLE'] to STRUCT2", "root missing field: NOT_NULLABLE"],
        )


if __name__ == "__main__":
    unittest.main()



import string
import random
from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructType
from datalakebundle.table.upsert.UpsertQueryCreator import UpsertQueryCreator
from datalakebundle.delta.DeltaStorage import DeltaStorage


class DataWriter:
    def __init__(
        self,
        spark: SparkSession,
        delta_storage: DeltaStorage,
        upsert_query_creator: UpsertQueryCreator,
    ):
        self.__spark = spark
        self.__delta_storage = delta_storage
        self.__upsert_query_creator = upsert_query_creator

    def append(self, df: DataFrame, full_table_name: str, schema: StructType, options: dict):
        df.select([field.name for field in schema.fields]).write.mode("append").options(**options).saveAsTable(full_table_name)

    def overwrite(self, df: DataFrame, full_table_name: str, partition_by: list, options: dict):
        self.__delta_storage.overwrite_data(df, full_table_name, partition_by, options)

    def upsert(self, df: DataFrame, full_table_name: str, schema: StructType, primary_key: list):
        temp_source_table = (
            f"upsert_{full_table_name.replace('.', '__')}_{''.join(random.choice(string.ascii_lowercase) for _ in range(6))}"
        )

        df.createOrReplaceTempView(temp_source_table)

        upsert_sql_statement = self.__upsert_query_creator.create(full_table_name, schema, primary_key, temp_source_table)

        try:
            self.__spark.sql(upsert_sql_statement)

        except BaseException:  # pylint: disable = broad-except, try-except-raise
            raise

        finally:
            self.__spark.catalog.dropTempView(temp_source_table)
